# local
try:
    from ts3ekkosingle.event_type import EventType
except ImportError:
    from .event_type import EventType


class Command:
    """
    Lower priority of the command means lower priority in the order the commands are called.
    If a command successfully handles an event, no further command will be called for that event.
    """

    def __init__(self, event_hooks, func, priority=None, name=None, help_cmd_str=None):
        """

        :param event_hooks: single or list of EventHooks this command should trigger on
        :param func: callback for event handler
        :param priority: higher=gets called earlier, handling stops on first successful handle
        :param name:
        :param help_cmd_str: Command prefix that should be displayed in the help command list.
        """
        if name is None:
            name = func.__name__

        if type(event_hooks) == EventType:
            event_hooks = [event_hooks]
        self.event_hooks = event_hooks

        self.func = func
        self.name = name

        self.help_cmd_str = help_cmd_str

        if priority is None:
            priority = 100
        self.priority = priority

    def __repr__(self):
        return f'<EkkoCmd name: {self.name}, prio: {self.priority}, ehooks: {self.event_hooks}>'



import argparse
import logging

# to provide gettext for the overwritten methods from NonExitParser, because the native argparse does so.
try:
    from gettext import gettext as _
except ImportError:
    def _(message):
        return message

# local
try:
    from ts3ekkosingle.errors import EkkoParsingError, EkkoArgparserMessage
except ImportError:
    from .errors import EkkoParsingError, EkkoArgparserMessage

class NonExitParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        if message:
            raise EkkoParsingError(message, self.format_usage())

    def error(self, message):
        args = {'prog': self.prog, 'message': message}
        self.exit(2, _('%(prog)s: error: %(message)s\n') % args)

    def print_help(self, file=None):
        """
        Catch the help generation and raise the flowcontrol EkkoArgparserMessage instead.
        """
        raise EkkoArgparserMessage(self.format_help())

    def print_usage(self, file=None):
        """
        Catch the usage generation and raise the flowcontrol EkkoArgparserMessage instead.
        """
        raise EkkoArgparserMessage(self.format_usage())


class MediaAliasParser:
    @staticmethod
    def parse_queue(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        parser.add_argument('--position', '-p', type=int)
        parser.add_argument('uri', nargs='+')
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return args.position, args.uri

    @staticmethod
    def parse_skip(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        parser.add_argument('count', default=1, nargs='?', type=int)
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return args.count

    @staticmethod
    def parse_volume(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        parser.add_argument('percentage', type=int)
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return args.percentage

    @staticmethod
    def parse_media(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        parser.add_argument('query_type', nargs='?', type=str)
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return args.query_type or None

class UtilityParser:
    @staticmethod
    def parse_name(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        parser.add_argument('name', nargs='+')
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return ' '.join(args.name)

    @staticmethod
    def parse_noargs(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return args

class FunParser:
    @staticmethod
    def parse_force_dice(cmd_prefix, cmd_str, **kwargs):
        parser = NonExitParser(prog=cmd_prefix, **kwargs)
        parser.add_argument('--force', '-f', action='store_true')
        parser.add_argument('choices', nargs='+')
        args = parser.parse_args(cmd_str[len(cmd_prefix):].split())
        return args.force, args.string



import asyncio
import ts3
import re
import time
import logging


# local
try:
    from ts3ekkosingle.errors import TS3AuthenticationFailed, TS3InsufficientPermissions, TS3InvalidCommand, \
        EkkoUnsuitedCommand, EkkoParsingError, EkkoArgparserMessage
    from ts3ekkosingle.command import Command
    from ts3ekkosingle.event_type import EventType
except ImportError:
    from .errors import TS3AuthenticationFailed, TS3InsufficientPermissions, TS3InvalidCommand, EkkoUnsuitedCommand, \
        EkkoParsingError, EkkoArgparserMessage
    from .command import Command
    from .event_type import EventType

logger = logging.getLogger('ekkoclient-ts3ekko')


class ProbeResult:
    SUCCESS = 1
    UNSUITED = 2
    PARSING_ERROR = 3
    CONN_RESET = 4
    PARSING_ABORT = 5


class TS3Ekko:
    def __init__(self, apikey, server='localhost', port='25639', loop=None, bot_username=None):
        self.server = server
        self.port = port

        self.apikey = apikey
        logger.debug('apikey=' + apikey)

        self.ts3conn = self.create_connection()
        self.bot_username = bot_username
        self.loop = loop or asyncio.get_event_loop()

        self.commands = []

    def add_command(self, hooks, func, **kwargs):
        """
        Add a command to the list of enabled commands and sort the list according to command priority.
        
        :param hooks: List of EventTypes on which the command should be called
        :param func: callback func for the command
        :param kwargs: various parameter for Command() creation
        """
        cmd = Command(hooks, func, **kwargs)

        self.commands.append(cmd)
        self.commands.sort(key=lambda cmd: cmd.priority)
        self.commands.reverse()

    def add_cog(self, cog):
        """
        Adds a cog to the bot. A cog is a collection of commands. The commands should be stored in the `.command`
         attribute in the following form:

         list of tuple in format: (EventType-int, callback-func, priority-int, name-str, help_cmd_str-str)

        For more information about these command configuration see :class:`Command`.
        :param cog:
        :return:
        """
        logger.debug(f'adding cog: {cog}')
        for cmd in cog.commands:
            try:
                self.add_command(cmd[0], cmd[1], priority=cmd[2], name=cmd[3], help_cmd_str=cmd[4])
            except Exception as e:
                logger.error(f'failed to add cog-command to collection: {cmd},'
                             f'\nexception: \n{e}')

    def create_connection(self, timeout=20):
        """
        Create a ts3query connection and authenticate with the apikey.
        
        :param timeout: Number of re-tries on failure.
        :return: authenticated ts3query connection
        """
        for to in range(0, timeout):
            try:
                _ts3conn = ts3.query.TS3ClientConnection(self.server, self.port)
                _ts3conn.auth(apikey=self.apikey)
                logger.debug(f'created TS3ClientQuery to {self.server}:{self.port}')
                return _ts3conn
            except OSError as e:
                logger.debug(f'failed to create TS3ClientQuery connection (os failed): {e}')
                time.sleep(1)
            except ts3.query.TS3QueryError as e:
                logger.debug(f'failed to create TS3ClientQuery connection (auth failed): {e.resp.error["msg"]}')
                time.sleep(1)
        logger.critical('failed to create TS3ClientQuery connection - TIMEOUT REACHED!')
        raise TS3AuthenticationFailed()

    async def poll_events(self):
        """
        Poll the ts3query connection for all possible events. Dispatch handling-process on event receive.
        """
        self.ts3conn.clientnotifyregister(event="any", schandlerid=0)
        while True:
            logger.debug('polling new events')
            self.ts3conn.send_keepalive()
            try:
                ts3_event = await self.loop.run_in_executor(None, self.ts3conn.wait_for_event, 60)
            except ts3.query.TS3TimeoutError:
                pass
            except Exception as e:
                logger.critical(f'while polling for event: {e}')
            else:
                logger.debug('found an event')
                # do all the work here and block here as well
                # all work on query has to be finished before the polling starts again
                try:
                    await self.handle_event(ts3_event)
                except Exception as e:
                    logger.critical(e)

    def _safe_handle_probe(self, cmd: Command, ts3event):
        """
        Probe and execute a Command if the command matches the received event.
        
        :param cmd: a ts3ekko command to be probed and executed if matching to event
        :param ts3event: a received ts3query event
        :return: ProbeResult
        """
        try:
            logger.debug(f'matching hook found, attempt handling: {cmd}')
            cmd.func(ts3event)
            # event got handled successfully
            # stop searching for handlers
            logger.debug(f'success handling')
            return ProbeResult.SUCCESS
        except EkkoUnsuitedCommand:
            # handler did not want the event or failed to handle it
            # continue probing other handlers
            logger.debug(f'failure handling')
            return ProbeResult.UNSUITED
        except EkkoParsingError:
            # Parser failed to parse the event successfully
            return ProbeResult.PARSING_ERROR
        except EkkoArgparserMessage:
            # Parser successfully parsed event but user requested help or usage information
            return ProbeResult.PARSING_ABORT
        except Exception as f:
            logger.critical(f)
            # Sometimes (for whatever reason) the connection is just not good
            logger.warning(f'resetting clientquery connection ...')
            self.ts3conn = self.create_connection()
            self.ts3conn.clientnotifyregister(event="any", schandlerid=0)
            return ProbeResult.CONN_RESET

    async def handle_event(self, ts3event):
        """
        Process a received event and dispatch the associated/matching command.
        :param ts3event: a received ts3query event
        """
        if ts3event[0].get('invokername', '') == self.bot_username:
            logger.debug(f"filtered message {ts3event[0]} (self sourced)")
            return
        # self.commands is sorted by priority already
        logger.debug(f'event arrived: {ts3event.__dict__}')
        logger.debug(f'current available commands: {self.commands}')
        logger.debug(f'event type: {EventType.get_type(ts3event)}')
        for cmd in self.commands:
            # check if the command hooks on the present ts3event
            if EventType.get_type(ts3event) in cmd.event_hooks:
                probe_result = self._safe_handle_probe(cmd, ts3event)
                if probe_result == ProbeResult.CONN_RESET:
                    if self._safe_handle_probe(cmd, ts3event) == ProbeResult.SUCCESS:
                        # went wrong on first try, but corrected after connreset
                        break
                    else:
                        # went wrong even after connreset
                        break
                elif probe_result == ProbeResult.SUCCESS:
                    # success on first try
                    break
                elif probe_result == ProbeResult.PARSING_ERROR:
                    # matching command was not able to parse the input correctly, abort!
                    break
                elif probe_result == ProbeResult.PARSING_ABORT:
                    # matching command was successfully parsed, but user requested help or usage information
                    break
                else:
                    # not a suited command
                    continue

    async def periodic_update(self, delay: int = 10):
        """
        A coroutine which performs updates/calls functions in a periodic manner. 
        
        It currently does not have any features implemented, other than a keep-alive/debug whoami call. 
        
        :param delay: number of seconds between updates.
        """

        def create_tconn():
            logger.debug('entered periodic update')
            temp_conn = self.create_connection()
            temp_conn.clientnotifyregister(event='any', schandlerid=0)
            logger.debug('created temporary p-u connection')
            return temp_conn

        temp_conn = create_tconn()

        while True:
            temp_conn.send_keepalive()
            try:
                logger.debug('PERIODIC_UPDATE_WHOAMI: ' + str(temp_conn.whoami()[0]))
            except Exception as f:
                logger.critical(f)
                temp_conn = create_tconn()
            await asyncio.sleep(delay)
            logger.debug('p-u sleep done')

    @staticmethod
    def check_cmd_suitability(regex, text, match: bool = False):
        """
        Wrapper for re.match, checks if a regex matches a given text. 
        Raises EkkoUnsuitedCommand on failure to find or to match regex.
        
        :param regex: regular expression to match/find
        :param text: text for regex match/find
        :param match: if regex should be matched (True) or only searched (False)
        :return: True if regex matched/was found and the command was deemed suitable.
        """
        if match:
            if not re.match(regex, text):
                raise EkkoUnsuitedCommand()
        else:
            if not re.search(regex, text):
                raise EkkoUnsuitedCommand()
        # if nothing got raised yet, then the command is suitable
        return True

    def start(self):
        """
        Initiates event polling and periodic updates. 
        """
        self.loop.run_until_complete(
            asyncio.wait([
                self.poll_events(),
                self.periodic_update(),
            ])
        )



import ts3


class TS3InvalidCommand(ts3.TS3Error):
    def __init__(self, cmd=None):
        self.cmd = cmd

    def __str__(self):
        if self.cmd is not None:
            return 'The command "{}" is not a valid command.'.format(self.cmd)
        else:
            return 'The command could not be parsed from the message (probably no command in message)'


class TS3InsufficientPermissions(ts3.TS3Error):
    def __init__(self, client=None, action=None):
        self.client = client
        self.action = action

    def __str__(self):
        return 'The client "{}" is not allowed to perform the action "{}"'.format(self.client, self.action)


class TS3AuthenticationFailed(ts3.TS3Error):
    def __str__(self):
        return 'Authentication failed - maybe endpoint is not up yet?'


class EkkoUnsuitedCommand(Exception):
    def __str__(self):
        return 'This command did not get processed by the handler because it was found unsuited.'


class EkkoNonexistentAlias(Exception):
    def __init__(self, aliasname, permanent=False):
        self.aliasname = aliasname
        self.permanent = permanent

    def __str__(self):
        return f'The alias \'{self.aliasname}\' (permanent={self.permanent}) does not exist.'


class EkkoNonexistentPermissionDoc(Exception):
    def __init__(self, permission):
        self.permission = permission

    def __str__(self):
        return f'There is no documentation for the permission \'{self.permission}\' available.'


class EkkoNonexistentGrant(Exception):
    def __init__(self, grant_id):
        self.grant_id = grant_id

    def __str__(self):
        return f'There is no permission grant with the id {self.grant_id}.'


class PermissionDenied(Exception):
    def __init__(self, msg: str, is_deny_request: bool):
        self.msg = msg
        self.__is_deny_request = is_deny_request

    @property
    def request_type(self) -> str:
        if self.__is_deny_request:
            return 'deny request'
        else:
            return 'grant request'

    def __repr__(self):
        return f'Permission denied, ({self.request_type}): {self.msg}'


class PermissionDuplicate(Exception):
    def __init__(self, grant):
        self.grant = grant

    def __repr__(self):
        return f'Permission Duplicate found for grant={self.grant}'


class EkkoParsingError(Exception):
    """
    To be raised by a custom argparser, to indicate a problem with the parsed input.
    """

    def __init__(self, message, usage):
        self.message = message
        self.usage = usage

    def __str__(self):
        return f'{self.usage}{self.message}'


class EkkoArgparserMessage(Exception):
    """
    Raised by a custom argparser, to indicate that the user requested either help or usage information 
    (which are included in this exception).
    """
    pass


class EkkoInvalidLocalMedia(Exception):
    """
    To be raised if a local media file is requested but can not be delivered, because it is not in the limiting
    media directory path or because it does not exist.
    """

    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return f'Access to file \'{self.filename}\' denied. Does not exist or you are not permitted to access it.'



#!/usr/bin/python3.6
import subprocess
import time
import os
import urllib.parse
import logging
import pathlib

import ts3ekkoutil.parser
import ts3ekkosingle
from ts3ekkoutil.envconsts import EkkoPropertyNames as epn

try:
    from ts3ekkosingle.ts3identity import replace_identity, startup
    from ts3ekkosingle.bot import EkkoBot
except ImportError:
    from .ts3identity import replace_identity, startup
    from .bot import EkkoBot

logger = logging.getLogger('ekkoclient-entrypoint')


def prepare_platform(prep_env):
    """
    Start services required for teamspeak. Uses sleeps to wait for required services to start (15 sec).

    :param prep_env: Set of environment variables, including DISPLAY=:1.
    """
    logger.info('starting xvfb')
    subprocess.Popen(['Xvfb', ':1', '-screen', '0', '1025x768x16'], env=prep_env) # Xvfb :1 -screen 0 1025x768x16
    time.sleep(5)
    logger.info('starting dbus')
    subprocess.run(['dbus-daemon', '--system', '--fork'], env=prep_env) # dbus-daemon --system --fork
    time.sleep(5)
    logger.info('starting pulseaudio')
    subprocess.run(['pulseaudio', '--start', '--daemonize=true'], env=prep_env)
    time.sleep(5)


def prepare_teamspeak(path, record_key, old_identity_str, new_identity_str):
    """
    Provision the ts3 client. Modifies the ts3 settings.db to put in the desired identity.

    :param path: path of the settings.db to be modified
    :param record_key: key-id of the record to be modified inside of the settings.db ProtobufItems table
    :param old_identity_str: string to be removed from the value of the targeted record (default id)
    :param new_identity_str: string to be inserted into the value of the targeted record (new id)
    """
    sess = startup(path)
    logger.debug(f'path: {path}, record key: {record_key}, '
                 f'old_iden: {old_identity_str}, new_iden: {new_identity_str}')
    replace_identity(sess, record_key, old_identity_str, new_identity_str)
    logger.info('ts3 identity prepared')


def start_teamspeak(prep_env, ts3_workdir, ts3_runscript, ts3url):
    """
    Start the teamspeak client.
    """
    ts3path = pathlib.Path(ts3_workdir) / ts3_runscript
    ts3path = ts3path.resolve()
    logger.info(f'starting teamspeak client in {ts3path}')
    subprocess.Popen([str(ts3path), ts3url], env=prep_env)


def assemble_ts3serverurl(server, port, username, server_password=None, channel_name=None, channel_id=None,
                          channel_password=None, token=None):
    """

    ts3server://host?port=9987&nickname=UserNickname&password=serverPassword&channel=MyDefaultChannel
        &cid=channelID&channelpassword=defaultChannelPassword&token=TokenKey&addbookmark=MyBookMarkLabel

    :param server: ip or url of ts3 server
    :param port: port of ts3 server
    :param username: wanted username
    :param server_password: password for the ts3 server
    :param channel_name: channel name to be connected to
    :param channel_id: channel id to be connected to (takes priority over name)
    :param channel_password: password for the channel to be connected to
    :param token: permission token
    :return: the assembled ts3 url
    """
    ts3url = f"ts3server://{server}?port={port}&nickname={urllib.parse.quote(username)}"

    if server_password is not None:
        ts3url += f"&password={urllib.parse.quote(server_password)}"

    if channel_name is not None:
        ts3url += f"&channel={urllib.parse.quote(channel_name)}"

    if channel_id is not None:
        ts3url += f"&cid={channel_id}"

    if channel_password is not None:
        ts3url += f"&channelpassword={urllib.parse.quote(channel_password)}"

    if token is not None:
        ts3url += f"&token={urllib.parse.quote(token)}"

    logger.debug(ts3url)
    return ts3url


def start_ekko(args, prep_env):
    """
    Start the ekko bot.
    :param args: 
    :return: 
    """
    logger.info('starting ekko')
    bot = EkkoBot(args=args, prep_env=prep_env)
    bot.start()
    logger.info('ekko started')


def main():
    parser = ts3ekkoutil.parser.create_ekko_parser()
    args = vars(parser.parse_args())

    logging.basicConfig(level=int(args[epn.LOG_LEVEL]), format=args[epn.LOG_FORMAT])

    logger.info(f'ts3ekkosingle version: {ts3ekkosingle.__version__}')

    # Set DISPLAY env variable for dbus, pulseaudio and teamspeak xvfb
    prep_env = os.environ.copy()
    prep_env["DISPLAY"] = ':1'

    logger.debug(args)
    logger.debug(os.environ)

    # Start services and client
    prepare_platform(prep_env)
    ts3url = assemble_ts3serverurl(args[epn.TS3_SERVER_HOST], args[epn.TS3_SERVER_PORT], args[epn.TS3_USERNAME],
                                   args[epn.TS3_SERVER_PASSWORD], args[epn.TS3_CHANNEL_NAME], args[epn.TS3_CHANNEL_ID],
                                   args[epn.TS3_CHANNEL_PASSWORD], args[epn.TS3_SERVER_PERMISSION_TOKEN])

    # make some basics checks about things that definitely need to be configured
    if args[epn.TS3_IDENTITY_DATABASE_KEY] is None:
        logger.error(f'{epn.TS3_IDENTITY_DATABASE_KEY} must be configured')

    if args[epn.TS3_IDENTITY_DATABASE_PRECONF] is None or args[epn.TS3_IDENTITY] is None:
        logger.error(f'{epn.TS3_IDENTITY_DATABASE_PRECONF} and {epn.TS3_IDENTITY} must be configured')

    prepare_teamspeak(f'{args[epn.TS3_CONFIG_DIRECTORY]}settings.db', args[epn.TS3_IDENTITY_DATABASE_KEY],
                      args[epn.TS3_IDENTITY_DATABASE_PRECONF], args[epn.TS3_IDENTITY])
    start_teamspeak(prep_env, args[epn.TS3_DIRECTORY], args[epn.TS3_RUNSCRIPT], ts3url)
    start_ekko(args, prep_env)


if __name__ == '__main__':
    main()



import logging
import random

from ts3ekkoutil.envconsts import EkkoPropertyNames as epn

try:
    from ts3ekkosingle.event_type import EventType
    from ts3ekkosingle.errors import EkkoUnsuitedCommand
    from ts3ekkosingle.argparser import NonExitParser, FunParser
except ImportError:
    from ..event_type import EventType
    from ..errors import EkkoUnsuitedCommand
    from ..argparser import NonExitParser,  FunParser

class FunPermission:
    DICE = 'fun.dice'

class FunCog:
    def __init__(self, ekkobot, args):
        self.ekko_bot = ekkobot


    @property
    def commands(self):
        """
        Returns data for command hooks creation for events like mediaalias set/get, etc.

        :return list of tuple in format: (EventType-int, callback-func, priority-int, name-str, help_cmd_str-str)
        """
        return [
            (EventType.TEXTMESSAGE, self.cmd_dice, 100, None, '!dice'),
        ]


    def cmd_dice(self, event):
        """
        Command: !dice

        :param event: TS3Event
        """
        cmd_prefix = '!dice'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(FunPermission.DICE, event):
                force, string = self.ekko_bot.parse(FunParser.parse_force_dice, event, cmd_prefix, event[0]['msg'])
                if not force:
                    self.ekko_bot.reply("Who do you think I am, the Taninator ? ", event)
                else:
                    choices = string.split(',')
                    self.ekko_bot.reply(f"The die rolls and lands on {random.choice(choices)}.", event)


import logging
import yaml

try:
    from ts3ekkosingle.errors import EkkoNonexistentPermissionDoc, EkkoNonexistentGrant, PermissionDenied, \
        PermissionDuplicate
except ImportError:
    from .errors import EkkoNonexistentPermissionDoc, EkkoNonexistentGrant, PermissionDenied, PermissionDuplicate

logger = logging.getLogger('ekkobot-permission')


class InvokerCtx:
    """
    Data collection about a specific user. Used to compare against other InvokerCtxs (e.g. permission restrictions)
     to determine if permission rules are matching or not.
    """

    def __init__(self, server_groups, channel_group, unique_id, username=None):
        """
        Create a new InvokerCtx object.

        :param server_groups: list of server group ids
        :param channel_group: channel group id
        :param unique_id: unique id
        :param username: (optional) username, will not be used for matching/comparing
        """
        self.server_groups = server_groups
        self.channel_group = channel_group
        self.unique_id = unique_id
        self.username = username

        if self.server_groups is not None:
            self.server_groups = [int(sgid) for sgid in self.server_groups]
        else:
            self.server_groups = []

        if self.channel_group is not None:
            self.channel_group = int(self.channel_group)

    def match(self, po):
        """
        Determine if a PermissionGrant or InvokerCtx matches with this InvokerCtx. Raises PermissionDenied if any attribute differs.
        
        :param po: PermissionGrant/InvokerCtx
        :raises PermissionDenied: on mismatch
        """
        if self.channel_group != po.channel_group and po.channel_group is not None:
            raise PermissionDenied('client_group does not match', False)
        if self.unique_id != po.unique_id and po.unique_id is not None:
            raise PermissionDenied('unique_id does not match', False)
        for server_group_id in po.server_groups or []:
            if server_group_id not in self.server_groups:
                raise PermissionDenied('server_groups do not match', False)

    def not_match(self, po):
        """
        Determine if a PermissionGrant or InvokerCtx does not match with this InvokerCtx. Raises PermissionDenied if no attribute differs.
        
        :param po: PermissionGrant/InvokerCtx
        :raises PermissionDenied: on full match.
        """
        if self.channel_group == po.channel_group or po.channel_group is None:
            if self.unique_id == po.unique_id or po.unique_id is None:
                for server_group_id in po.server_groups or []:
                    if server_group_id not in self.server_groups:
                        return
                raise PermissionDenied('client_group, unique_id and server_groups match', True)

    def __repr__(self):
        repr = f'<ts3ekkosingle.permission.InvokerCtx server_groups:"{self.server_groups}", ' \
               f'channel_group:"{self.channel_group}", unique_id: "{self.unique_id}"'
        if self.username is not None:
            repr += f', username: "{self.username}"'
        repr += '>'
        return repr


class PermissionManager:
    """

        Sort of thinking about permissions in yaml style.
        A set of servergroups, channelgroups and identities is constructed with AND clauses.
        All sets together are compared with OR.

        The master permission (which allows access to all commands) is called "master".

        ---
        - master:
            - name: xyoz
              identity:
                32dk3okljwe/32xsr3&23ex==

        - media.queue.append:
            -   name: generals
                servergroups:
                    1
                    2
                    24
                channelgroups:
                    2
                    3
                    4
                identity:
                    vm3405789j34kew23/r23x==
            -   name: friendos
                servergroups:
                    3
        - media.queue.skip:
            -   name: idiots
                servergroups:
                    2

            [etc...]
    """
    def __init__(self, path):
        self.permissions = {}
        with open(path, 'r') as f:
            yaml_permissions = yaml.safe_load(f)

        for action, sets in yaml_permissions.items():
            for perm_set in sets:
                try:
                    self.permissions[action].append(
                        InvokerCtx(perm_set.get('servergroups'), perm_set.get('channelgroup'), perm_set.get('identity'))
                    )
                except KeyError:
                    self.permissions[action] = [
                        InvokerCtx(perm_set.get('servergroups'), perm_set.get('channelgroup'), perm_set.get('identity'))
                    ]

    def _query_grants(self, action: str, ictx: InvokerCtx):
        """
        Query database to find a matching grant to the given action/permission and InvokerCtx which would allow 
        the invoker to perform the requested action.
        
        If this function executes without PermissionDenied being raised, then a matching grant was found.
        
        This is the counterpart to `_query_denies`. Only if both functions are called with the same parameters 
        and do not raise the PermissionDenied error, the invoker is actually allowed for the given permission.
        
        :param action: requested permission name
        :param ictx: InvokerCtx about the invoker that requested the permission
        :raises PermissionDenied: if no matching grant could be found
        """
        for result in self.permissions.get(action, []):
            try:
                ictx.match(result)
                logger.info(f'permission finally granted - ictx: {ictx} for action: {action}')
                return
            except PermissionDenied as e:
                logger.debug(e)
                # denied (for this rule, need to check all other rules as well)
                pass
        logger.info(f'permission finally denied - ictx: {ictx} for action: {action}')
        raise PermissionDenied('', False)


    def can(self, action: str, context: InvokerCtx) -> bool:
        """
        Checks if an action/permission is allowed in the given invoker context.
        
        :param action: requested permission name
        :param context: invoker identifications
        :return: bool depending if the action is allowed (True) or denied (False)
        """

        try:
            self._query_grants('master', context)
        except PermissionDenied:
            pass
        else:
            return True

        try:
            self._query_grants(action, context)
        except PermissionDenied:
            return False
        else:
            return True



import logging
import mpv
import youtube_dl
import datetime
import re
import pathlib

from ts3ekkoutil.envconsts import EkkoPropertyNames as epn

try:
    from ts3ekkosingle.event_type import EventType
    from ts3ekkosingle.errors import EkkoUnsuitedCommand, EkkoNonexistentAlias, EkkoInvalidLocalMedia
    from ts3ekkosingle.argparser import NonExitParser, MediaAliasParser, UtilityParser
except ImportError:
    from ..event_type import EventType
    from ..errors import EkkoUnsuitedCommand, EkkoNonexistentAlias, EkkoInvalidLocalMedia
    from ..argparser import NonExitParser, MediaAliasParser, UtilityParser

logger = logging.getLogger('ekkocog-mediaalias')


class MediaCogPermission:
    QUEUE = 'media.queue.append'
    SKIP = 'media.queue.skip'
    MEDIA = 'media.queue.media'
    MEDIA_QUEUE = 'media.queue.media_queue'
    CLEAR_QUEUE = 'media.queue.clear'
    SHUFFLE_QUEUE = 'media.queue.shuffle'
    PAUSE = 'media.pause'
    RESUME = 'media.resume'

    VOLUME_SET = 'media.volume.set'
    VOLUME_GET = 'media.volume.get'
    VOLUME_RESET = 'media.volume.reset'

    HOHOHO = 'media.hohoho'

class TrackMeta:
    def __init__(self, uri):
        self._uri = uri
        self.info_dict = None

    def _ensure_info_dict(self):
        if self.info_dict is None:
            logger.debug(f'dict_info for uri={self.uri} not a available, downloading')
            try:
                with youtube_dl.YoutubeDL() as ytdl:
                    self.info_dict = ytdl.extract_info(self.uri, download=False)
                    logger.debug(self.info_dict)
            except:
                self.info_dict = {}

    @property
    def uri(self):
        if self._uri.startswith('ytdl://'):
            return f'https://youtu.be/{self._uri[len("ytdl://"):]}'
        else:
            return self._uri

    @property
    def title(self):
        self._ensure_info_dict()
        return self.info_dict.get('title', self.uri)

    @property
    def bbcode_formatted(self):
        if self.is_local_file:
            return self.uri
        else:
            return f'[url={self.uri}]{self.title}[/url]'

    @property
    def is_local_file(self):
        return not self.uri.startswith('https://') and not self.uri.startswith('http://')


class MediaCog:
    def __init__(self, ekkobot, args):
        self.ekko_bot = ekkobot

        self._volume_modifier = args[epn.COG_MEDIA_VOLUME_MODIFIER]
        self._volume_max = args[epn.COG_MEDIA_VOLUME_MAX]

        self.media_directory = args[epn.EKKO_MEDIA_DIRECTORY]

        # create mpv player
        self.mpv = mpv.MPV(ytdl=True)
        # this is a music bot, not a video bot! no need to spend resources on video.
        self.mpv.vid = False
        # callback to remove played tracks from mpv playlist
        self.mpv.event_callback('end-file')(self._remove_played_track)

        # cache for track metadata (title, etc)
        self.trackmeta_cache = {}

        # Init volume at a 'safe' level
        self._reset_volume()
        # Init hohoho volume
        self._hohoho_volume = self.mpv.volume * 1.7
        self._hohoho_uri = "https://www.youtube.com/watch?v=8lFjrAj-vrk"

    @property
    def commands(self):
        """
        Returns data for command hooks creation for events like mediaalias set/get, etc.
        
        :return list of tuple in format: (EventType-int, callback-func, priority-int, name-str, help_cmd_str-str)
        """
        return [
            (EventType.TEXTMESSAGE, self.cmd_queue, 100, None, '!queue'),
            (EventType.TEXTMESSAGE, self.cmd_skip, 100, None, '!skip'),
            (EventType.TEXTMESSAGE, self.cmd_media, 100, None, '!media'),
            (EventType.TEXTMESSAGE, self.cmd_pausemedia, 100, None, '!pausemedia'),
            (EventType.TEXTMESSAGE, self.cmd_resumemedia, 100, None, '!resumemedia'),
            (EventType.TEXTMESSAGE, self.cmd_clearqueue, 100, None, '!clearqueue'),
            (EventType.TEXTMESSAGE, self.cmd_shuffle, 100, None, '!shuffle'),
            (EventType.TEXTMESSAGE, self.cmd_volume, 75, None, '!volume'),
            (EventType.TEXTMESSAGE, self.cmd_volume_set, 100, None, '!volume'),
            (EventType.TEXTMESSAGE, self.cmd_volume_get, 100, None, '!volume'),
            (EventType.TEXTMESSAGE, self.cmd_volumereset, 100, None, '!volume'),
            (EventType.TEXTMESSAGE, self.cmd_hohoho, 100, None, '!hohoho'),
            (EventType.TEXTMESSAGE, self.cmd_hohoho_volume, 100, None, '!hohoho-volume'),
            (EventType.TEXTMESSAGE, self.cmd_hohoho_uri, 100, None, '!hohoho-uri'),
        ]

    @staticmethod
    def _remove_url_bbcode(dirty):
        """
        Removes the `[URL]` and `[/URL]` BB-Code from a string.
        
        :param dirty: string or list to be cleared
        :return: string or list without the URL BB-Code
        """
        if type(dirty) == str:
            return re.sub("\[\/?[uU][rR][lL]\]", '', dirty)
        elif type(dirty) == list:
            return [re.sub("\[\/?[uU][rR][lL]\]", '', uri) for uri in dirty]

    def _reset_volume(self):
        """
        Resets the mpv volume.
        """
        self.mpv.volume = (self._volume_max * self._volume_modifier) * 0.5

    def _remove_played_track(self, event):
        """
        Remove played tracks from mpv playlist.
        
        :param event: the mpv event which creates the call to this function
        """
        logger.debug(f'_remove_played_track {event}')
        logger.debug(f'_remove_played_track pre-removal playlist: {self.mpv.playlist}')
        logger.debug(f'_remove_played_track pre-removal playlist_pos: {self.mpv.playlist_pos}')
        for index in range(0, self.mpv.playlist_pos or 1):
            if self.mpv.playlist and not self.mpv.playlist[0].get('current', False):
                self.mpv.playlist_remove(0)
                logger.debug(f'_remove_played_track post-removal playlist: {self.mpv.playlist}')
                logger.debug(f'_remove_played_track post-removal playlist_pos: {self.mpv.playlist_pos}')

    def _media_info_queue(self, end=None):
        """
        Creates and formats text-output of all currently queued tracks.
        
        :param end: maximum amount of listed tracks
        :return: bb-code formatted output string
        """
        output = 'Currently queued tracks:\n'
        for index in range(self.mpv.playlist_pos or 0, end or len(self.mpv.playlist or [])):
            try:
                uri = self.mpv.playlist[index]['filename']
                trackmeta = self._request_trackmeta(uri)
                output += f'{trackmeta.bbcode_formatted}\n'
            except IndexError:
                break
        if self.mpv.playlist_pos is None:
            output += 'No track queued or playing!'
        return output.strip('\n')

    def _media_info_current(self):
        """
        Creates and formats text-output for the currently playing track.
        
        :return: bb-code formatted output string
        """
        if self.mpv.playlist_pos is not None:
            uri = self.mpv.playlist[self.mpv.playlist_pos]['filename']
            trackmeta = self._request_trackmeta(uri)
            try:
                td_remaining = datetime.timedelta(seconds=self.mpv.playtime_remaining)
                formatted_remaining = self.format_seconds(td_remaining.seconds)
                td_played = datetime.timedelta(seconds=self.mpv.playback_time)
                formatted_played = self.format_seconds(td_played.seconds)
            except TypeError:
                return f'{trackmeta.bbcode_formatted}'
            else:
                return f'{trackmeta.bbcode_formatted} - {formatted_played} played, {formatted_remaining} remaining'
        else:
            return 'No track queued or playing!'

    @staticmethod
    def format_seconds(seconds: int):
        """
        Formats an amount of seconds into a mm:ss format. If in hour-range, also add hours in front (h:mm:ss)
        
        :param seconds: amount of seconds to be formatted
        :type seconds: int
        :return: string in h:mm:ss format.
        """
        result = ''
        if seconds // 3600 > 0:
            result += f'{seconds//3600}:'
        result += f'{str(seconds%3600//60).zfill(2)}:{str(seconds%60).zfill(2)}'
        return result

    @staticmethod
    def resolve_mediafile_path(media_directory, file):
        """
        Resolves relative path/file name inside the configured media directory.
        
        :param media_directory: media directory root path
        :param file: file name of the to be resolved file
        :return: absolute path of the requested file
        :raises: EkkoInvalidLocalMedia on invalid file name
        """
        media_directory_path = pathlib.Path(media_directory).resolve()
        # TODO: yeah, would be nice to have glob support for local fs | Problem: single uri -> multiple uri
        # media_paths = [mp.resolve for mp in media_directory_path.glob(file)]
        media_path = media_directory_path / file
        media_path = media_path.resolve()

        # check if resolved path is also in the configured media directory and actually exists
        # => prevent fs index creation/info acquirement
        for fragment in media_path.parents:
            if fragment == media_directory_path and media_path.exists():
                return str(media_path)

        # apparently it violates something (existing/in jail), so lets raise an eception
        raise EkkoInvalidLocalMedia(file)

    def _request_trackmeta(self, uri: str):
        """
        Provides TrackMeta object for requested track. Tries to fetch from cache first, otherwise creates.

        :param uri: requested track
        :return: TrackMeta object
        """
        trackmeta = self.trackmeta_cache.get(uri, None)

        if trackmeta is None:
            # not in cache, lets create and cache it!
            trackmeta = TrackMeta(uri)
            self.trackmeta_cache[uri] = trackmeta

        return trackmeta

    def playlist_append(self, uris, pos):
        """
        Inserts tracks into the requested position in the mpv playlist.
        Adjusts mpv playlist_pos accordingly if necessary.

        Should multiple tracks be inserted the pos param marks the position of the first insert and
        others will be inserted afterwards.

        :param uris: list of uris to be inserted into the playlist.
        :param pos: position where to tracks should be inserted. None = append.
        """

        for index, uri in enumerate(uris):
            uri = self._remove_url_bbcode(uri)

            if TrackMeta(uri).is_local_file:
                try:
                    uri = self.resolve_mediafile_path(self.media_directory, uri)
                except EkkoInvalidLocalMedia as e:
                    logger.error(e)
                    continue

            logger.debug(f'adding "{uri}" to playlist')

            self.mpv.playlist_append(uri)

            if self.mpv.playlist_pos is None:
                logger.debug('playlist_pos is none, setting to 0')
                self.mpv.playlist_pos = 0

            if pos is not None:
                logger.debug('pos is not none, modifying playlist order')
                last_index = len(self.mpv.playlist) - 1
                self.mpv.playlist_move(last_index, self.mpv.playlist_pos + pos + index)

                # if new pos is 0, fix the playlist_pos
                if pos == 0:
                    self.mpv.playlist_pos = 0

        logger.debug(self.mpv.playlist)

    def cmd_skip(self, event):
        """
        Command: !skip

        Skips a given number of tracks (default 1) in the playlist.

        See `MediaAliasParser.parse_skip` for parameters.

        :param event: TS3Event
        """
        cmd_prefix = '!skip'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.SKIP, event):
                count = self.ekko_bot.parse(MediaAliasParser.parse_skip, event, cmd_prefix, event[0]['msg'])
                playlist_length = len(self.mpv.playlist or [])
                if playlist_length <= count:
                    # lib does not have a method for stop, so lets access mpv more directly.
                    self.mpv.command('stop')
                else:
                    self.mpv.playlist_pos += count

    def cmd_queue(self, event):
        """
        Command: !queue

        Queues a given track into the playlist.

        See `MediaAliasParser.parse_queue` for parameters.

        :param event: TS3Event
        """
        cmd_prefix = '!queue'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix} ', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.QUEUE, event):
                pos, uris = self.ekko_bot.parse(MediaAliasParser.parse_queue, event, cmd_prefix, event[0]['msg'])
                uris = self._remove_url_bbcode(uris)
                self.playlist_append(uris, pos)

    def cmd_media(self, event):
        """
        Command: !media

        Queries metadata about media in the playlist.

        See `MediaAliasParser.parse_media` for parameters.

        :param event: TS3Event
        """
        cmd_prefix = '!media'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}(\s*$|\s+\w+)', event[0]['msg']):
            query_type = self.ekko_bot.parse(MediaAliasParser.parse_media, event, cmd_prefix, event[0]['msg'],
                                             description='Show information about current media. '
                                                         'If `queue` keyword is specifed, show information about '
                                                         'the next few queued tracks.')
            if query_type == 'queue':
                if self.ekko_bot.can(MediaCogPermission.MEDIA_QUEUE, event):
                    self.ekko_bot.reply(self._media_info_queue(5), event)
            else:
                if self.ekko_bot.can(MediaCogPermission.MEDIA, event):
                    self.ekko_bot.reply(self._media_info_current(), event)

    def cmd_pausemedia(self, event):
        """
        Command: !pausemedia

        Pauses media playback.

        :param event: TS3Event
        """
        cmd_prefix = '!pausemedia'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.PAUSE, event):
                self.ekko_bot.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                                    description='Pause the current playing media.')
                self.mpv.pause = True
                self.ekko_bot.reply('Playback paused. Resume with "!resumemedia".', event)

    def cmd_resumemedia(self, event):
        """
        Command: !resumemedia

        Resumes media playback.

        :param event: TS3Event
        """
        cmd_prefix = '!resumemedia'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.RESUME, event):
                self.ekko_bot.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                                    description='Resume playing the previously paused media.')
                self.mpv.pause = False
                self.ekko_bot.reply('Playback resumed. Pause with "!pausemedia".', event)

    def cmd_clearqueue(self, event):
        """
        Command: !clearqueue

        Removes all not-playing tracks from the queue. Does not stop the current playing track.

        :param event: TS3Event
        """
        cmd_prefix = '!clearqueue'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.CLEAR_QUEUE, event):
                self.ekko_bot.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                                    description='Clear the playlist queue and remove all currently not playing'
                                                ' tracks from it.')
                self.mpv.playlist_clear()
                self.ekko_bot.reply('Queue cleared.', event)

    def cmd_shuffle(self, event):
        """
        Command: !shuffle

        Shuffles the current playlist.

        :param event: TS3Event
        """
        cmd_prefix = '!shuffle'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.SHUFFLE_QUEUE, event):
                self.ekko_bot.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                                    description='Shuffles all currently queued tracks.')
                self.mpv.command('playlist-shuffle')
                # After playlist-shuffle, the current playing track will not be found at playlist_pos=0 anymore, but will
                # have a new position inside the playlist. To not have all tracks between playlist_pos=0 and current
                # playlist_pos removed by `_remove_played_track`, we quickly want to put back the current playing song
                # at position 0.
                self.mpv.playlist_move(self.mpv.playlist_pos, 0)
                self.ekko_bot.reply('Queue shuffled.', event)

    def cmd_volumereset(self, event):
        """
        Command: !volume reset

        Resets the volume to the default value.

        :param event: TS3Event
        """
        cmd_prefix = '!volume reset'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.VOLUME_RESET, event):
                self.ekko_bot.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                                    description='Reset the volume to the default level.')
                self._reset_volume()
                self.ekko_bot.reply('Volume reset.', event)

    def cmd_volume(self, event):
        """
        Command: !volume [-h|--help]

        Fallback command in case none of the other !volume commands catches on, replies with usage information.

        :param event: TS3Event
        """
        cmd_prefix = '!volume'
        cmd_usage = 'usage: !volume [new volume] \n    get current volume: !volume\n    set new volume: !volume 100'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}\s*(-h|--help)\s*$', event[0]['msg']):
            self.ekko_bot.reply(cmd_usage, event)

    def cmd_volume_get(self, event):
        """
        Command: !volume

        Replies with current volume value.

        :param event: TS3Event
        """
        cmd_prefix = '!volume'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}\s*$', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.VOLUME_GET, event):
                self.ekko_bot.reply(f'Current volume: {self.mpv.volume / self._volume_modifier}', event)

    def cmd_volume_set(self, event):
        """
        Command: !volume <value>

        Sets the volume to the given value.

        :param event: TS3Event
        """
        cmd_prefix = '!volume'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix} \d+\s*$', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.VOLUME_SET, event):
                volume = self.ekko_bot.parse(MediaAliasParser.parse_volume, event, cmd_prefix, event[0]['msg'])
                if volume < self._volume_max:
                    self.mpv.volume = volume * self._volume_modifier
                else:
                    self.mpv.volume = self._volume_max * self._volume_modifier
                self.ekko_bot.reply(f'Volume set to {self.mpv.volume / self._volume_modifier}.', event)

    def cmd_hohoho(self, event):
        """
        Command: !hohoho

        Plays a hearty christmas-y santa ho-ho-ho~!

        :param event: TS3Event
        """
        cmd_prefix = '!hohoho'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix}\s*$', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.HOHOHO, event):
                # create mpv player
                hohoh_mpv = mpv.MPV(ytdl=True)
                # this is a music bot, not a video bot! no need to spend resources on video.
                hohoh_mpv.vid = False
                # set volume
                hohoh_mpv.volume = self._hohoho_volume
                hohoh_mpv.playlist_append("https://www.youtube.com/watch?v=zajl0ZGHbAk")

    def cmd_hohoho_volume(self, event):
        """
        Command: !hohoho-volume <value>

        Sets the hohoho volume to the given value.

        :param event: TS3Event
        """
        cmd_prefix = '!hohoho-volume'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix} \d+\s*$', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.HOHOHO, event):
                volume = self.ekko_bot.parse(MediaAliasParser.parse_volume, event, cmd_prefix, event[0]['msg'])
                if volume < self._volume_max:
                    self._hohoho_volume = volume * self._volume_modifier
                else:
                    self._hohoho_volume = self._volume_max * self._volume_modifier
                self.ekko_bot.reply(f'hohoho-volume set to {self._hohoho_volume / self._volume_modifier}.', event)

    def cmd_hohoho_uri(self, event):
        """
        Command: !hohoho-uri <value>

        Sets the hohoho uri to the given value.

        :param event: TS3Event
        """
        cmd_prefix = '!hohoho-uri'
        if self.ekko_bot.check_cmd_suitability(f'^{cmd_prefix} \d+\s*$', event[0]['msg']):
            if self.ekko_bot.can(MediaCogPermission.HOHOHO, event):
                pos, uris = self.ekko_bot.parse(MediaAliasParser.parse_queue, event, cmd_prefix, event[0]['msg'])
                uris = self._remove_url_bbcode(uris)
                if type(uris) == list:
                    uris = uris[0]
                self._hohoho_uri = uris
                self.ekko_bot.reply(f'hohoho-uri set to {self._hohoho_uri}.', event)



from sqlalchemy import Column, Integer, create_engine, VARCHAR
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
import hashlib
import logging

Base = declarative_base()


class ProtobufItems(Base):
    __tablename__ = 'ProtobufItems'
    timestamp = Column(Integer, nullable=False)
    key = Column(VARCHAR, nullable=False, unique=True, primary_key=True)
    value = Column(VARCHAR)


def startup(path_to_settings_db):
    engine = create_engine(f'sqlite:///{path_to_settings_db}')

    Base.metadata.bind = engine

    DBSession = sessionmaker(bind=engine)

    session = DBSession()
    return session


def replace_identity(sess, key, old_identity, new_identity):
    """
    Replaces the private identity in the record identified by the key.
    :param sess: sqalchemy session
    :param key: identifier of the record to be edited
    :param old_identity: string which should be replaced
    :param new_identity: replacement string
    """
    if type(old_identity) == str:
        logging.debug('encoded old identity to bytes')
        old_identity = old_identity.encode('latin1')

    if type(new_identity) == str:
        logging.debug('encoded new identity to bytes')
        new_identity = new_identity.encode('latin1')

    identity_query = sess.query(ProtobufItems).filter_by(key=key)
    identity_record = identity_query.one()
    identity_query.update({'value': identity_record.value.replace(old_identity, new_identity)})
    sess.commit()
    checksum_update(sess)
    logging.debug('identity thing should be done')


def checksum_update(sess):
    """
    Recalculates the Checksum record in the ProtobufItems table in the given sqlalchemy session
    :param sess: sqlalchemy session
    """
    identity_query = sess.query(ProtobufItems).filter(ProtobufItems.key != 'Checksum')
    joined_pool = b''.join([iden.value for iden in identity_query])
    new_checksum = hashlib.sha1(joined_pool).digest()

    checksum_query = sess.query(ProtobufItems).filter_by(key='Checksum')
    logging.debug(sess.query(ProtobufItems).filter_by(key='Checksum').one().value)
    checksum_query.update({'value': new_checksum})
    logging.debug(sess.query(ProtobufItems).filter_by(key='Checksum').one().value)
    sess.commit()

    logging.debug(sess.query(ProtobufItems).filter_by(key='Checksum').one().value)


if __name__ == '__main__':
    old_identity = '15456318Vrjp+cffeguX6meh5OMc6vEcNT51xCFp1awUMZVFiWnYiWXl4A1dGCFYACyZeJn9SL1NLWl9hAz8iUn0GcFNEeUEIaXdwLVUIOS8JPQ8WVz9GBlthIm1BOAIvR1N0F2B0DV9Sdk1sa0NJUURUTGhjbnZHQnU4QWUzbXBEQkhRMVBKeDUrc2RuYXFmYXFQOUVwM0lQbXdnPT0='
    new_identity = '34951128VxRulcsSPzgOgHnjUisP10QWbuoJ7LkRkclsocCFLYngnfkNZdUhHE1FWKA5RKkx6GQF3Vnhif1NWcQwie39EcAhGAWFnLn1eVwhDCQwAJBZjH296NFdAI0IJWXx6BFlgIXEEV2x3ZW9BaUVBNEcwa0VVSGltQndWdEtkZTd2SVY4bm01N2pBcXB6Yzh3YmdYcGw5V0w1bz0='
    #new_identity = '68103251VWij5a8B5ezKXh+lQreyptqbPg8xfCFB+X3c7BhYaXVI1WV5AW1kFBiFDARZ2UGNlDVN6eUpYalY0Qn0GVnF+SlhGCEpnA3V2Dw94V1APIj0FXFBUJ3p/CVNXUE5CL1VJDghXQk5Ra0NJUUNKVVJoYis1cmd1YUpKUjZvNG51QUhyZjlZMUN5S0VIMXJPQXRXeVdBNjhnPT0='
    sess = startup('/home/xyoz/.ts3client/settings.db')
    replace_identity(sess, 4, old_identity, new_identity)



import enum


class EventType(enum.Enum):
    """
    Enum of existing/available ts3 event types.
    """
    UNKNOWN = 0

    TEXTMESSAGE = 1
    TALKSTATUSCHANGE = 2
    MESSAGE = 3
    MESSAGELIST = 4
    COMPLAINLIST = 5
    BANLIST = 6
    CLIENTMOVED = 7
    CLIENTLEFTVIEW = 8
    CLIENTENTERVIEW = 9
    CLIENTPOKE = 10
    CLIENTCHATCLOSED = 11
    CLIENTCHATCOMPOSING = 12
    CLIENTUPDATED = 13
    CLIENTIDS = 15
    CONNECTIONINFO = 16
    CHANNELCREATED = 17
    CHANNELEDITED = 18
    CHANNELDELETED = 19
    CHANNELMOVED = 20
    SERVEREDITED = 21
    SERVERUPDATED = 22
    CURRENTSERVERCONNECTIONCHANGED = 23
    CONNECTSTATUSCHANGE = 24
    SERVERGROUPLIST = 25

    @classmethod
    def get_type(cls, ts3event):
        if ts3event.event == 'notifytextmessage':
            return EventType.TEXTMESSAGE
        elif ts3event.event == 'notifytalkstatuschange':
            return EventType.TALKSTATUSCHANGE
        elif ts3event.event == 'notifymessage':
            return EventType.MESSAGE
        elif ts3event.event == 'notifymessagelist':
            return EventType.MESSAGELIST
        elif ts3event.event == 'notifycomplainlist':
            return EventType.COMPLAINLIST
        elif ts3event.event == 'notifybanlist':
            return EventType.BANLIST
        elif ts3event.event == 'notifyclientmoved':
            return EventType.CLIENTMOVED
        elif ts3event.event == 'notifyclientleftview':
            return EventType.CLIENTLEFTVIEW
        elif ts3event.event == 'notifycliententerview':
            return EventType.CLIENTENTERVIEW
        elif ts3event.event == 'notifyclientpoke':
            return EventType.CLIENTPOKE
        elif ts3event.event == 'notifyclientchatclosed':
            return EventType.CLIENTCHATCLOSED
        elif ts3event.event == 'notifyclientchatcomposing':
            return EventType.CLIENTCHATCOMPOSING
        elif ts3event.event == 'notifyclientupdated':
            return EventType.CLIENTUPDATED
        elif ts3event.event == 'notifyconnectioninfo':
            return EventType.CONNECTIONINFO
        elif ts3event.event == 'notifychannelcreated':
            return EventType.CHANNELCREATED
        elif ts3event.event == 'notifychanneledited':
            return EventType.CHANNELEDITED
        elif ts3event.event == 'notifychanneldeleted':
            return EventType.CHANNELDELETED
        elif ts3event.event == 'notifychannelmoved':
            return EventType.CHANNELMOVED
        elif ts3event.event == 'notifyserveredited':
            return EventType.SERVEREDITED
        elif ts3event.event == 'notifyserverupdated':
            return EventType.SERVERUPDATED
        elif ts3event.event == 'notifycurrentserverconnectionchanged':
            return EventType.CURRENTSERVERCONNECTIONCHANGED
        elif ts3event.event == 'notifyconnectstatuschange':
            return EventType.CONNECTSTATUSCHANGE
        elif ts3event.event == 'notifyservergrouplist':
            return EventType.SERVERGROUPLIST
        else:
            return EventType.UNKNOWN




import logging
import subprocess
import ts3.response
from ts3ekkoutil.envconsts import EkkoPropertyNames as epn

try:
    from ts3ekkosingle.ts3ekko import TS3Ekko
    from ts3ekkosingle.cogs.media import MediaCog
    from ts3ekkosingle.cogs.fun import FunCog
    from ts3ekkosingle.event_type import EventType
    from ts3ekkosingle.errors import EkkoUnsuitedCommand, EkkoParsingError, EkkoNonexistentPermissionDoc, \
        EkkoArgparserMessage, EkkoNonexistentGrant
    from ts3ekkosingle.permission import InvokerCtx, PermissionManager
    from ts3ekkosingle.argparser import UtilityParser
except ImportError:
    from .ts3ekko import TS3Ekko
    from .cogs.media import MediaCog
    from .cogs.fun import FunCog
    from .event_type import EventType
    from .errors import EkkoUnsuitedCommand, EkkoParsingError, EkkoNonexistentPermissionDoc, EkkoArgparserMessage, \
        EkkoNonexistentGrant
    from .permission import InvokerCtx, PermissionManager
    from .argparser import UtilityParser

logger = logging.getLogger('ekkoclient-ekkobot')


class EkkoCorePermission:
    JOIN = 'control.join'
    NAME = 'control.name'
    PERMISSION_RELOAD = 'control.permission_reload'
    WHOAMI = 'other.whoami'
    MUTEFIX = 'other.mutefix'


class EkkoBot(TS3Ekko):
    def __init__(self, args, prep_env):
        self.args = args
        self._prep_env = prep_env
        super().__init__(apikey=args[epn.TS3_CLIENT_APIKEY], server=args[epn.TS3_CLIENTQUERY_HOST],
                         port=args[epn.TS3_CLIENTQUERY_PORT], bot_username=args[epn.TS3_USERNAME])

        self.register_hooks()
        self.permission_manager = PermissionManager(args[epn.EKKO_PERMISSION_PATH])

        self.add_cog(MediaCog(self, self.args))
        self.add_cog(FunCog(self, self.args))

    def can(self, permission: str, event: ts3.response.TS3Event, quiet=False,
            deny_msg='Sorry, can\'t do that. You might not have the required permission for that.',
            quiet_permission_name=False, invoker_ctx: InvokerCtx = None) -> bool:
        """
        Can/Permission-Check relay with invoker specification to permission manager

        :param permission: name of the required permission
        :param event: TS3Event from which the required invoker information can be read from
        :param quiet: if True, the method wont reply to the event with a "denied" message in case of denied permission
        :param deny_msg: content of message that should be sent out when the action/permission was denied
        :param quiet_permission_name: if True, the method wont mention the required permission in case of denied permission
        :param invoker_ctx: if specified, use the specified invoker context instead of determining it from the passed event
        :return: bool, if the invoker is allowed to perform the action or not
        """
        if invoker_ctx is None:
            invoker_ctx = self.parse_invoker_context(event[0])
        if not self.permission_manager.can(permission, invoker_ctx):
            if not quiet and quiet_permission_name:
                self.reply(deny_msg, event)
            elif not quiet and not quiet_permission_name:
                self.reply(f'{deny_msg} (permission name: {permission})', event)
            return False
        else:
            return True

    def parse(self, parse_func, event: ts3.response.TS3Event, *args, **kwargs):
        """
        Calls a parsing function with the given arguments. Replies with usage and error in teamspeak to the event,
        should the parsing be not successful.

        :param parse_func: a argparser function (see in `~ts3ekkosingle.parser`)
        :param event: a ts3 event to which can be replied
        :param args: args for the parser function
        :param kwargs: kwargs for the parser function
        :returns: from the parse_func
        :raises EkkoParsingError: if parse_func is not successful
        :raises EkkoArgparserMessage: if user requested help/usage information
        """
        try:
            return parse_func(*args, **kwargs)
        except EkkoParsingError as e:
            # Something was missing (required args e.g.)
            self.reply(e.__str__(), event)
            raise
        except EkkoArgparserMessage as e:
            # These are messages like help or usage
            self.reply(e.__str__(), event)
            # This exception should be excepted further up the handling chain and be quietly discarded,
            # while also aborting the handling (because it was handled succesfully).
            raise

    def register_hooks(self):
        """
        Enables command hooks for events like spawning and despawning.
        FIXME: would really like this to be decorators, guess I could do that sometimes in the future
        """
        self.add_command(EventType.TEXTMESSAGE, self.cmd_join, help_cmd_str='!join')
        self.add_command(EventType.TEXTMESSAGE, self.cmd_help)
        self.add_command(EventType.TEXTMESSAGE, self.cmd_name, help_cmd_str='!name')
        self.add_command(EventType.TEXTMESSAGE, self.cmd_whoami, help_cmd_str='!whoami')
        self.add_command(EventType.TEXTMESSAGE, self.cmd_mutefix, help_cmd_str='!mutefix')
        self.add_command(EventType.TEXTMESSAGE, self.cmd_reloadpermissions, help_cmd_str='!reloadperm')
        self.add_command(EventType.CONNECTSTATUSCHANGE, self.init_description)

    def init_description(self, event):
        """
        Not yet implemented due to design fault/bug in the ClientQuery API.
        => changing own description requires dbproperties-edit permission, which you usually dont have.
        
        See: https://forum.teamspeak.com/threads/134743-ClientQuery-edit-own-description
        """
        if event[0]['status'] == 'connection_established':
            pass
        raise EkkoUnsuitedCommand()

    def cmd_join(self, event):
        """
        Command: !join

        Moves the bot instance into the channel the invoker of the command is in. 
        
        # TODO: add support for channel passwords

        :param event: TS3Event
        """
        cmd_prefix = '!join'
        if self.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.can(EkkoCorePermission.JOIN, event):
                self.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                           description='Make this bot instance join your channel.')
                invoker_clid = event[0]['invokerid']
                invoker_cid = self.ts3conn.cid_from_clid(invoker_clid)
                bot_clid = self.ts3conn.whoami()[0]['clid']
                self.ts3conn.clientmove(clid=bot_clid, cid=invoker_cid)

    def cmd_name(self, event):
        """
        Command: !name

        Renames this bot instance to the given name.
        
        See `UtilityParser.parse_name` for parameters.

        :param event: TS3Event
        """
        cmd_prefix = '!name'
        if self.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.can(EkkoCorePermission.NAME, event):
                name = self.parse(UtilityParser.parse_name, event, '!name', event[0]['msg'],
                                  description='Change the bot\'s username')
                self.ts3conn.clientupdate(client_nickname=name)

    def cmd_whoami(self, event):
        """
        Command: !whoami

        Replies technical information about the invokers teamspeak identity on this server.

        :param event: TS3Event
        """
        cmd_prefix = '!whoami'
        if self.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.can(EkkoCorePermission.WHOAMI, event):
                self.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                           description='Get technical information about your teamspeak identity on this server.')
                ictx = self.parse_invoker_context(event[0])
                self.reply(f'username: {ictx.username} | server groups: {ictx.server_groups} | '
                           f'channel group: {ictx.channel_group} | unique id: {ictx.unique_id}', event)
    def cmd_mutefix(self, event):
        """
        Command: !mutefix

        Mutes the pulseaudio speaker on which teamspeak continues to send notification sounds despite being turned off.

        :param event: TS3Event
        """
        cmd_prefix = '!mutefix'
        if self.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.can(EkkoCorePermission.MUTEFIX, event):
                self.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                           description='Mutes the notification sounds that the bot echos.')
                subprocess.run(['pactl', 'set-sink-input-mute', '0', '1'], env=self._prep_env)

    def cmd_reloadpermissions(self, event):
        """
        Command: !reloadpermissions

        Re-creates the permission manager which forces a reload of the permission file.

        :param event: TS3Event
        """
        cmd_prefix = '!reloadperm'
        if self.check_cmd_suitability(f'^{cmd_prefix}', event[0]['msg']):
            if self.can(EkkoCorePermission.PERMISSION_RELOAD, event):
                self.parse(UtilityParser.parse_noargs, event, cmd_prefix, event[0]['msg'],
                           description='Reloads the permissions.')
                self.permission_manager = PermissionManager(self.args[epn.EKKO_PERMISSION_PATH])

    def cmd_help(self, event):
        """
        Command: !help

        Lists all publicly available commands.

        :param event: TS3Event
        """
        if self.check_cmd_suitability('^!help', event[0]['msg']):
            help_cmd_strs = sorted(set(
                [str(cmd.help_cmd_str) for cmd in self.commands if cmd.help_cmd_str is not None]
            ))
            help_reply = f'List of all available command sets: \n    ' + '\n    '.join(help_cmd_strs) + \
                         '\n To receive more help about a specific command, use the `-h` or `--help` ' \
                         'flag when calling a command (e.g. !skip --help)'
            self.reply(help_reply, event)

    @staticmethod
    def color_message(text, color='#1433b1'):
        """
        Puts BB-Code formatting around the provided text.

        :param text: text which should be colored.
        :param color: the color which should be put on the text.
        :return: color BB-code formatted string.
        """
        return f'[color={color}]{text}[/color]'

    @staticmethod
    def remove_ts3_smilies(text):
        """
        Place a zero-width space inbetween characters which are transformed on display into ts3 smilies.
        
        Does not replace ':/' as this messes up the linking of [url=https://...][/url].
        
        :param text: text to be replaced in
        :return: text, but with disabled smilies
        """
        return str(text) \
            .replace(':)', ':\u200B)') \
            .replace(':D', ':\u200BD') \
            .replace('8)', '8\u200B)') \
            .replace(';)', ';\u200B)') \
            .replace(':(', ':\u200B(') \
            .replace(':C', ':\u200BC') \
            .replace(':0', ':\u200B0') \
            .replace(':x', ':\u200Bx') \
            .replace(':P', ':\u200BP')

    def resolve_servergroups(self) -> dict:
        """
        Create dataset of servergroup-ids (sgids) and their respective string name.

        FIXME: doesnt work, pls fix (ts3conn.servergrouplist() returns nothing/returns later on different notify)
        :return: dict => keys are the sgids, the data is the string name
        """
        resolved_sgids = {}
        try:
            sgid_dataset = self.ts3conn.servergrouplist()
            logger.debug(sgid_dataset.__dict__)
            for sgid_data in sgid_dataset:
                resolved_sgids[str(sgid_data['sgid'])] = sgid_data['name']
        except IndexError:
            pass
        logger.debug(resolved_sgids)
        return resolved_sgids

    def parse_invoker_context(self, event) -> InvokerCtx:
        """
        Creates InvokerCtx (Invoker Context) based on the data available in the provided event. This includes
        data like server & channel groups of the invoker, as well as their unique id.

        :param event: TS3Event
        :return: InvokerCtx()
        """
        client_id = event['invokerid']

        client_variables = self.ts3conn.clientvariable(client_id, "client_channel_group_id", "client_servergroups")[0]
        logger.debug(client_variables)
        return InvokerCtx(client_variables['client_servergroups'].split(','),
                          client_variables['client_channel_group_id'], event['invokeruid'], event['invokername'])

    @staticmethod
    def split_message(remaining_message, part_length=1024, split_chars=('\n', ' ', ''), text_mod=None):
        """
        Splits a message into parts to match a certain maximum length restriction.

        :param remaining_message: string/message which should be split
        :param part_length: maximum length of each part
        :param split_chars: characters used to split the message into parts, FIFO processing
        :param text_mod: function which transforms text (like color coding, needed to match part_length)
        :return: list of parts from the message
        """

        def dummy_text_mod(s):
            return s

        if text_mod is None:
            text_mod = dummy_text_mod

        continue_pad = '\n'
        parts = []
        # process split_chars until all parts match part_length
        for split_char in split_chars:
            while True:
                if len(remaining_message) <= 0:
                    return parts

                # split message and prepare for re-assemble
                # Also: str.split() doesnt really like to split on empty separators (like '')
                # so we need to handle that separately
                if split_char != '':
                    split_result = remaining_message.split(split_char)
                else:
                    split_result = list(remaining_message)

                stitched_message = ''
                for splitter in split_result:
                    # create imaginary part based on the current splitter and transform it in the text_mod function
                    future_part = text_mod(f'{continue_pad}{stitched_message}{splitter}{split_char}')
                    # check if this imaginary part would meet the requirement
                    if len(future_part) <= part_length:
                        stitched_message += splitter + split_char
                    else:
                        break

                if not parts and stitched_message != '':
                    # the first part should not have the continue_pad
                    # remove last appended split char again
                    if split_char != '':
                        parts.append(text_mod(stitched_message[:-len(split_char)]))
                    else:
                        # splicing with a len==0 character as length results in empty string. lets not do that.
                        parts.append(text_mod(stitched_message))
                elif stitched_message != '':
                    # remove last appended split char again
                    if split_char != '':
                        parts.append(text_mod(continue_pad + stitched_message[:-len(split_char)]))
                    else:
                        # splicing with a len==0 character as length results in empty string. lets not do that.
                        parts.append(text_mod(continue_pad + stitched_message))
                else:
                    break

                # remove processed string from remaining message
                remaining_message = remaining_message[len(stitched_message):]

    def reply(self, message, event):
        """
        Reply to the text message described by the event parameter.

        :param message: the bots message towards the other client.
        :param event: the source event to which should be replied to.
        :raises NotImplementedError: if the source event is not a text message.
        """
        if EventType.get_type(event) == EventType.TEXTMESSAGE:
            logger.debug(message)
            logger.debug(event[0])
            for message in self.split_message(message, part_length=1000, text_mod=lambda
                    s: self.color_message(self.remove_ts3_smilies(s))):
                self.ts3conn.sendtextmessage(targetmode=event[0]['targetmode'],
                                             target=event[0]['invokerid'],
                                             msg=message)
        else:
            raise NotImplementedError()
